
# THIS FILE IS GENERATED FROM UWGeodynamics SETUP.PY
#
short_version = '0.9.8b'
version = '0.9.8b'
full_version = '0.9.8b-dev-41bb885(master)'
git_revision = '41bb8852d31c1c2e6a27d684217bd2ade9f82e19'
release = False

if not release:
    version = full_version
